import com.tencent.subscribe.avro.MessageType;
import com.tencent.subscribe.avro.Operation;
import com.tencent.subscribe.avro.Record;
import org.apache.avro.specific.SpecificRecordBase;
import org.apache.commons.cli.*;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.header.Header;
import org.codehaus.jackson.map.ObjectMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.DatatypeConverter;
import java.nio.ByteBuffer;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.*;

public class SubscribeFromKafka {
    private static final Logger log = LoggerFactory.getLogger(SubscribeFromKafka.class);
    private static Map<Integer, partitionMsgConsumer> partitionMsgConsumers = new HashMap<>();
    private final static String partitionSeqKey = "ps";
    private final static String partitionSeq2Key = "ps2";
    private final static String CHECKPOINT = "Checkpoint";
    private final static String HEARTBEAT = "Heartbeat";

    public static void main(final String[] args) {
        Option option_brokers = Option.builder()
                .required(true)
                .desc("The brokers option")
                .longOpt("brokers")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_topic = Option.builder()
                .required(true)
                .desc("The topic option")
                .longOpt("topic")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_group = Option.builder()
                .required(true)
                .desc("The group option")
                .longOpt("group")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_user = Option.builder()
                .required(false)
                .desc("The user option")
                .longOpt("user")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_password = Option.builder()
                .required(false)
                .desc("The password option")
                .longOpt("password")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_trans = Option.builder()
                .required(false)
                .desc("The trans2sql option")
                .longOpt("trans2sql")
                .build();

        Options options = new Options()
                .addOption(option_brokers)
                .addOption(option_topic)
                .addOption(option_user)
                .addOption(option_password)
                .addOption(option_group)
                .addOption(option_trans);

        CommandLineParser parser = new DefaultParser();

        try {
            CommandLine commandLine = parser.parse(options, args);

            run(
                    commandLine.getOptionValue("brokers"),
                    commandLine.getOptionValue("topic"),
                    commandLine.getOptionValue("group"),
                    commandLine.getOptionValue("user"),
                    commandLine.getOptionValue("password"),
                    commandLine.hasOption("trans2sql")
            );

        } catch (Exception exception) {
            System.out.print("error: ");
            System.out.println(exception.getMessage());
        }
    }

    private static void run(
            String brokers,
            String topic,
            String group,
            String user,
            String password,
            boolean trans2sql) throws Exception {
        Properties props = new Properties();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, brokers);
        props.put(ConsumerConfig.GROUP_ID_CONFIG, group);
        props.put(ConsumerConfig.CLIENT_ID_CONFIG, group + "_" + Instant.now().getEpochSecond());
        props.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        props.put(ConsumerConfig.ISOLATION_LEVEL_CONFIG, "read_committed");
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
        props.put(ConsumerConfig.FETCH_MAX_BYTES_CONFIG, 20 << 20); // kafka消息上限20M
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.ByteArrayDeserializer");

        if (null != user && !user.equals("")) {
            props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_PLAINTEXT");
            props.put(SaslConfigs.SASL_MECHANISM, "SCRAM-SHA-512");
            props.put(SaslConfigs.SASL_JAAS_CONFIG,
                    "org.apache.kafka.common.security.scram.ScramLoginModule required " +
                            "  username=\"" + user + "\"" +
                            "  password=\"" + password + "\";");
        }

        KafkaConsumer<String, byte[]> consumer = new KafkaConsumer<>(props);
        consumer.subscribe(Collections.singletonList(topic), new partitionSubLis());
        try {
            while (true) {
                ConsumerRecords<String, byte[]> records = consumer.poll(Duration.ofSeconds(1));
                for (ConsumerRecord<String, byte[]> record : records) {
                    partitionMsgConsumer pc = partitionMsgConsumers.get(record.partition());
                    if (null == pc) {
                        throw new IllegalStateException("no such partition consumer " + record.partition());
                    }

                    pc.consume(consumer, record, trans2sql);
                }
            }
        } finally {
            consumer.close();
        }
    }


    private static class partitionMsgConsumer {
        long lastPS = 0;
        long lastPS2 = 0;
        TopicPartition tp;
        Record largeRecord;

        private partitionMsgConsumer(TopicPartition tp) {
            this.tp = tp;
        }

        private void consume(KafkaConsumer<String, byte[]> consumer, ConsumerRecord<String, byte[]> record, boolean trans2sql) throws Exception {
            // Each message has a sequence id in the partition, this id is continuous monotonic increasing,
            // these messages are duplicated if the id rotate;
            // There is unexpected message lost in kafka if the id is discontinuous,
            // have to reset the subscribe and client when it happened.
            // ps2是ps的升级版，2022.11.1之后启动的订阅任务包含此属性，可以最大限度地解决Kafka消息重复投递的问题，
            // ps解决的是Kafka分区中消息重复投递(Kafka自身机制导致)的问题，ps2在ps的基础上还可以解决后端任务重启后
            // 重复解析部分binlog导致的数据重复投递问题。
            long ps2 = getPartitionSeq2(record);
            long validPs;
            if (ps2 != -1) {
                if (lastPS2 > 0) {
                    if (ps2 < lastPS2) {
                        System.out.println("Duplicated message: " + record);
                        return;
                    } else if (ps2 == lastPS2) { // 可能是重复消息也可能是checkpoint或heartbeat消息
                        if (!isCheckpoint(record) && !isHeartbeat(record)) {
                            System.out.println("Duplicated message: " + record);
                            return;
                        }
                    } else {
                        if (ps2 != lastPS2 + 1) {
                            throw new IllegalStateException("the partition seq2 is discontinuous, last: " + lastPS2 + ", current: " + ps2);
                        }
                    }
                }
                lastPS2 = ps2;
                validPs = ps2;
            } else { // 老的订阅任务没有ps2，会走到这里
                long ps = getPartitionSeq(record);
                if (ps == -1) {

                } else if (ps <= lastPS) {
                    System.out.println("Duplicated message: " + record);
                    return;
                } else if (ps != lastPS + 1 && lastPS != 0) {
                    throw new IllegalStateException("the partition seq is discontinuous, last: " + lastPS + ", current: " + ps);
                }
                lastPS = ps;
                validPs = ps;
            }

            Record avroRecord = Deserializer.avroDeserialize(record.value());

            // version 2 需要考虑大消息合并，生产端对于超过10M的大消息做了拆分
            if (avroRecord.getVersion() == 2 && avroRecord.getTotal() > 1) {
                System.out.printf("merging large record, partition:%d, offset:%d, partitionSeq:%d, Total:%d, Index:%d, GTID:%s\n",
                        record.partition(), record.offset(), validPs, avroRecord.getTotal(), avroRecord.getIndex(), avroRecord.getGtid());
                if (avroRecord.getIndex() == 0) {
                    largeRecord = avroRecord;
                } else {
                    if (avroRecord.getOldColumnsBytes() != null) {
                        largeRecord.setOldColumnsBytes(concatBytes(largeRecord.getOldColumnsBytes(), avroRecord.getOldColumnsBytes()));
                    }
                    if (avroRecord.getNewColumnsBytes() != null) {
                        largeRecord.setNewColumnsBytes(concatBytes(largeRecord.getNewColumnsBytes(), avroRecord.getNewColumnsBytes()));
                    }
                }
                if (avroRecord.getIndex() < avroRecord.getTotal() - 1) {
                    return;
                }
                // 到这里就合并完成了，进行反序列化
                ObjectMapper objectMapper = new ObjectMapper();
                if (largeRecord.getOldColumnsBytes() != null) {
                    System.out.printf("largeRecord.OldColumnsBytes.length:%d, \n", largeRecord.getOldColumnsBytes().remaining());
                    if (largeRecord.getOldColumnsBytes().remaining() > 0) {
                        byte[] byteArray = new byte[largeRecord.getOldColumnsBytes().remaining()];
                        largeRecord.getOldColumnsBytes().get(byteArray);
                        Object mapData = objectMapper.readValue(byteArray, Object.class);
                        //组合后的大消息字段反序列化之后的格式和正常消息有一些差异，需要转一下格式
                        largeRecord.setBeforeImages(largeMsgAvroDataTransfer(mapData));
                    }
                    largeRecord.setOldColumnsBytes(null);
                }
                if (largeRecord.getNewColumnsBytes() != null) {
                    System.out.printf("largeRecord.NewColumnsBytes.length:%d, \n", largeRecord.getNewColumnsBytes().remaining());
                    if (largeRecord.getNewColumnsBytes().remaining() > 0) {
                        byte[] byteArray = new byte[largeRecord.getNewColumnsBytes().remaining()];
                        largeRecord.getNewColumnsBytes().get(byteArray);
                        Object mapData = objectMapper.readValue(byteArray, Object.class);
                        //组合后的大消息字段反序列化之后的格式和正常消息有一些差异，需要转一下格式
                        largeRecord.setAfterImages(largeMsgAvroDataTransfer(mapData));
                    }
                    largeRecord.setNewColumnsBytes(null);
                }
                avroRecord = largeRecord;
                avroRecord.setIndex(-1);
            }

            // 将腾讯云 avro Record 数据结构转换为 ali avro Record
            com.alibaba.dts.formats.avro.Record aliRecord = ClassConvertor.convert2AliAvroRecord(avroRecord);

            // 将 ali avro Record 传入业务处理函数做进一步业务处理，
            // 你可以在 onRecord 函数中自由扩展对 record 的处理逻辑，实现自己的业务逻辑。
            UserBusinessProcess.onRecord(record.partition(), record.offset(), validPs, aliRecord, trans2sql);

            // 提交Kafka消费位点，默认的提交机制是消费到 CHECKPOINT 消息做一次提交，
            // 在生产端一般每隔约 10s 会写入一个 CHECKPOINT 消息。
            if (avroRecord.getOperation() == Operation.CHECKPOINT) {
                consumer.commitAsync(Collections.singletonMap(tp, new OffsetAndMetadata(record.offset() + 1)), null);
            }
        }
    }


    private static long getPartitionSeq(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(partitionSeqKey).iterator();
        if (it.hasNext()) {
            Header h = it.next();
            return Long.parseUnsignedLong(new String(h.value()));
        }
        return -1;
    }

    private static long getPartitionSeq2(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(partitionSeq2Key).iterator();
        if (it.hasNext()) {
            Header h = it.next();
            return Long.parseUnsignedLong(new String(h.value()));
        }
        return -1;
    }

    private static boolean isCheckpoint(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(CHECKPOINT).iterator();
        return it.hasNext();
    }

    private static boolean isHeartbeat(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(HEARTBEAT).iterator();
        return it.hasNext();
    }


    /**
     * 针对超大消息中的oldColumnsBytes/newColumnsBytes字段，
     * 将反序列化后的Map结构的数据结构化为Avro类型的数据结构，
     * 以便使得大消息的数据格式与普通消息的数据格式保持一致，方便后续解析。
     **/
    public static Object largeMsgAvroDataTransfer(Object avroData) {
        try {
            if (avroData instanceof Map) {
                List<Object> colsList = new ArrayList<>();
                List<Map<String, Object>> array = (List<Map<String, Object>>) ((Map) avroData).get("array");
                for (Map<String, Object> avroMap : array) {
                    String avroType = (String) avroMap.keySet().toArray()[0];
                    Map<String, Object> innerMap = (Map<String, Object>) avroMap.get(avroType);
                    // create avro object
                    Class<?> avroClass = Class.forName("com.alibaba.dts.formats.avro." + avroType);
                    SpecificRecordBase avroObj = (SpecificRecordBase) avroClass.getDeclaredConstructor().newInstance();
                    for (String key : innerMap.keySet()) {
                        Object value = innerMap.get(key);
                        // 这几种类型原始数据为字节数组，序列化过程中自动做了base64转换，需要还原
                        if ((avroType.equals("Character") || avroType.equals("BinaryObject") || avroType.equals("BinaryGeometry"))
                                && "value".equals(key)) {
                            try {
                                byte[] decodedValue = Base64.getDecoder().decode((String) value);
                                avroObj.put(key, ByteBuffer.wrap(decodedValue));
                            } catch (Exception e) {
                                System.out.println("warning: No base64 data in Character/BinaryObject/BinaryGeometry. Exception: " + e);
                                // decode失败则忽略保存原值
                                avroObj.put(key, value);
                            }
                        } else {
                            avroObj.put(key, value);
                        }
                    }
                    colsList.add(avroObj);
                }
                return colsList;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return avroData;
    }

    private static ByteBuffer concatBytes(ByteBuffer a, ByteBuffer b) {
        if (a == null) {
            return b;
        }
        ByteBuffer mergedBuffer = ByteBuffer.allocate(a.remaining() + b.remaining());
        mergedBuffer.put(a);
        mergedBuffer.put(b);
        mergedBuffer.flip();
        return mergedBuffer;
    }

    private static class partitionSubLis implements ConsumerRebalanceListener {
        public void onPartitionsAssigned(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.computeIfAbsent(p.partition(), k -> new partitionMsgConsumer(p));
            }
        }

        public void onPartitionsRevoked(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.remove(p.partition());
            }
        }
    }
}
