import com.alibaba.dts.formats.avro.EmptyObject;
import com.alibaba.dts.formats.avro.Field;
import com.alibaba.dts.formats.avro.Operation;
import com.alibaba.dts.formats.avro.Record;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.DatatypeConverter;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

/**
 * UserBusinessProcess 类用于承载用户业务处理逻辑，这里仅定义了一个示例函数 onRecord，你可在此根据业务逻辑自由扩展自己的消费程序。
 **/
public class UserBusinessProcess {
    private static final Logger log = LoggerFactory.getLogger(UserBusinessProcess.class);

    /**
     * onRecord 函数仅对传入的 record 做打印或者转为 SQL 语句打印出来这样的处理，仅供参考。
     **/
    public static void onRecord(int partitionNum, long offset, long partitionSeq, Record record, boolean trans2sql) throws Exception {
        log.info("partition:" + partitionNum + ", offset:" + offset + ", partitionSeq:" + partitionSeq
                + ", " + record.toString());
        // 将 record 对象解析为 SQL 语句对很多业务并非需要的结果，请根据下游对数据格式的要求自行决定。
        // 这里的转换结果仅供参考，适用于常见 SQL 语句，无法支持各种 MySQL 版本的全部语法规则，如在线上系统使用请进行严格测试。
        if (trans2sql) {
            trans2SQL(partitionNum, offset, partitionSeq, record);
        }
    }


    private static void trans2SQL(int partitionNum, long offset, long partitionSeq, Record record) throws IllegalStateException {
        if (record.getOperation() != Operation.HEARTBEAT && record.getOperation() != Operation.CHECKPOINT) {
            log.info(
                    String.format("-->[partition: %d, offset: %d, partitionSeq: %d], happenedAt: %s",
                            partitionNum, offset, partitionSeq,
                            LocalDateTime.ofInstant(Instant.ofEpochSecond(record.getSourceTimestamp()),
                                    TimeZone.getDefault().toZoneId())));
        }

        switch (record.getOperation()) {
            case HEARTBEAT:
                System.out.println("heartbeat: " + record.getSourceTimestamp());
                break;
            case CHECKPOINT:
                Map<String, String> tags = record.getTags();
                System.out.println("Checkpoint: gtid=" + tags.get(ClassConvertor.TAG_GTID_SET)
                        + ", sourcePosition=" + record.getSourcePosition());
                break;
            case BEGIN:
                System.out.println("BEGIN");
                break;
            case COMMIT:
                System.out.println("COMMIT");
                break;
            case ROLLBACK:
                System.out.println("ROLLBACK");
                break;
            case DDL:
                System.out.println(record.getAfterImages());
                break;
            case INSERT:
                try {
                    String sql = transInsert(record);
                    System.out.println(sql);
                } catch (IllegalStateException e) {
                    e.printStackTrace();
                }
                break;
            case DELETE:
                try {
                    String sql = transDelete(record);
                    System.out.println(sql);
                } catch (IllegalStateException e) {
                    e.printStackTrace();
                }
                break;
            case UPDATE:
                try {
                    String sql = transUpdate(record);
                    System.out.println(sql);
                } catch (IllegalStateException e) {
                    e.printStackTrace();
                }
                break;
            default:
                throw new IllegalStateException("unsupported message type: " + record.getOperation());
        }
    }

    private static String transInsert(Record record) throws IllegalStateException {
        try {
            String objectName = record.getObjectName();
            String schemaName;
            String tableName;
            if (objectName.contains(".")) {
                schemaName = objectName.split("\\.")[0];
                tableName = objectName.split("\\.")[1];
            } else {
                throw new IllegalStateException("invalid objectName value: " + objectName);
            }
            StringBuilder sql = new StringBuilder(String.format("INSERT INTO `%s`.`%s` VALUES ", schemaName, tableName));
            if (!(record.getAfterImages() instanceof List)) {
                throw new IllegalStateException("unsupported data type of NewColumns: " + record.getAfterImages().getClass().getName());
            }
            List row = (List) record.getAfterImages();
            List<Field> columns = (List<Field>) record.getFields();
            String sep = "";
            sql.append('(');
            for (int i = 0; i < row.size(); i++) {
                sql.append(sep);
                sql.append(decode(columns.get(i), row.get(i)));
                sep = ", ";
            }
            sql.append(')');
            return sql.toString();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }


    private static String transDelete(Record record) throws IllegalStateException {
        try {
            String objectName = record.getObjectName();
            String schemaName;
            String tableName;
            if (objectName.contains(".")) {
                schemaName = objectName.split("\\.")[0];
                tableName = objectName.split("\\.")[1];
            } else {
                throw new IllegalStateException("invalid objectName value: " + objectName);
            }
            StringBuilder sql = new StringBuilder(String.format("DELETE FROM `%s`.`%s` WHERE ", schemaName, tableName));
            String sep = "";
            List<Field> columns = (List<Field>) record.getFields();
            if (!(record.getBeforeImages() instanceof List)) {
                throw new IllegalStateException("unsupported data type of OldColumns: " + record.getBeforeImages().getClass().getName());
            }
            List row = (List) record.getBeforeImages();
            for (int i = 0; i < row.size(); i++) {
                Field column = columns.get(i);
                Object value = row.get(i);
                if (null == value) {
                    continue;
                }

                sql.append(sep);
                sql.append(String.format("`%s`", column.getName()));
                if (value instanceof EmptyObject) {
                    sql.append(" IS NULL");
                } else {
                    sql.append(String.format(" = %s", decode(column, value)));
                }
                sep = " AND ";
            }
            sql.append(" LIMIT 1");
            return sql.toString();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    private static String transUpdate(Record record) throws IllegalStateException {
        try {
            String objectName = record.getObjectName();
            String schemaName;
            String tableName;
            if (objectName.contains(".")) {
                schemaName = objectName.split("\\.")[0];
                tableName = objectName.split("\\.")[1];
            } else {
                throw new IllegalStateException("invalid objectName value: " + objectName);
            }
            StringBuilder sql = new StringBuilder(String.format("UPDATE `%s`.`%s` SET ", schemaName, tableName));
            String sep = "";
            List<Field> columns = (List<Field>) record.getFields();
            if (!(record.getAfterImages() instanceof List)) {
                throw new IllegalStateException("unsupported data type of NewColumns: " + record.getAfterImages().getClass().getName());
            }
            List newRow = (List) record.getAfterImages();
            for (int i = 0; i < newRow.size(); i++) {
                Field column = columns.get(i);
                Object value = newRow.get(i);
                if (value == null) {
                    continue;
                }

                sql.append(sep);
                sql.append(String.format("`%s` = %s", column.getName(), decode(column, value)));
                sep = ", ";
            }

            if (!(record.getBeforeImages() instanceof List)) {
                throw new IllegalStateException("unsupported data type of OldColumns: " + record.getBeforeImages().getClass().getName());
            }
            List oldRow = (List) record.getBeforeImages();
            sql.append(" WHERE ");
            sep = "";
            for (int i = 0; i < oldRow.size(); i++) {
                Field column = columns.get(i);
                Object value = oldRow.get(i);
                if (null == value) {
                    continue;
                }

                sql.append(sep);
                sql.append(String.format("`%s`", column.getName()));
                if (value instanceof EmptyObject) {
                    sql.append(" IS NULL");
                } else {
                    sql.append(String.format(" = %s", decode(column, value)));
                }
                sep = " AND ";
            }
            sql.append(" LIMIT 1");
            return sql.toString();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    // 该方法不对value判空，需要提前针对不同类型的dml作判空处理
    private static String decode(Field field, Object value) {
        if (null == value) {
            return "DEFAULT";
        }
        Class<?> clz = value.getClass();
//        System.out.println("field.name: " + field.getName() + ", claasName: " + clz.getTypeName());
        if (com.alibaba.dts.formats.avro.Integer.class.equals(clz)) {
            return DataTypeParser.parseInteger((com.alibaba.dts.formats.avro.Integer) value);
        } else if (com.alibaba.dts.formats.avro.Character.class.equals(clz)) {
            try {
                return "_binary'" + DataTypeParser.parseCharacter((com.alibaba.dts.formats.avro.Character) value) + "'";
            } catch (Exception e) {
                return "x'" + DatatypeConverter.printHexBinary(((com.alibaba.dts.formats.avro.Character) value).getValue().array()) + "'";
            }
        } else if (com.alibaba.dts.formats.avro.Decimal.class.equals(clz)) {
            return DataTypeParser.parseDecimal((com.alibaba.dts.formats.avro.Decimal) value);
        } else if (com.alibaba.dts.formats.avro.Float.class.equals(clz)) {
            if (field.getDataTypeNumber() == DataTypeParser.MYSQL_TYPE_FLOAT) {
                return DataTypeParser.parseFloat((com.alibaba.dts.formats.avro.Float) value);
            }
            return DataTypeParser.parseDouble((com.alibaba.dts.formats.avro.Float) value);
        } else if (com.alibaba.dts.formats.avro.Timestamp.class.equals(clz)) {
            return "_binary'" + DataTypeParser.parseTimestamp((com.alibaba.dts.formats.avro.Timestamp) value) + "'";
        } else if (com.alibaba.dts.formats.avro.DateTime.class.equals(clz)) {
            return "_binary'" + DataTypeParser.parseDateTime(
                    (com.alibaba.dts.formats.avro.DateTime) value,
                    null,
                    field.getDataTypeNumber()) + "'";
        } else if (com.alibaba.dts.formats.avro.TimestampWithTimeZone.class.equals(clz)) {
            return "_binary'" + DataTypeParser.parseDateTime(
                    ((com.alibaba.dts.formats.avro.TimestampWithTimeZone) value).getValue(),
                    ((com.alibaba.dts.formats.avro.TimestampWithTimeZone) value).getTimezone(),
                    field.getDataTypeNumber()) + "'";
        } else if (com.alibaba.dts.formats.avro.BinaryGeometry.class.equals(clz)) {
            return "x'" + DataTypeParser.parseBinaryGeometry((com.alibaba.dts.formats.avro.BinaryGeometry) value) + "'";
        } else if (com.alibaba.dts.formats.avro.TextGeometry.class.equals(clz)) {
            return "_binary'" + DataTypeParser.parseTextGeometry((com.alibaba.dts.formats.avro.TextGeometry) value) + "'";
        } else if (com.alibaba.dts.formats.avro.BinaryObject.class.equals(clz)) {
            return "x'" + DataTypeParser.parseBinaryObject((com.alibaba.dts.formats.avro.BinaryObject) value) + "'";
        } else if (com.alibaba.dts.formats.avro.TextObject.class.equals(clz)) {
            if ("json".equals(((com.alibaba.dts.formats.avro.TextObject) value).getType())) {
                return "'" + DataTypeParser.parseTextObject((com.alibaba.dts.formats.avro.TextObject) value) + "'";
            } else {
                return "_binary'" + DataTypeParser.parseTextObject((com.alibaba.dts.formats.avro.TextObject) value) + "'";
            }
        } else if (com.alibaba.dts.formats.avro.EmptyObject.class.equals(clz)) {
            return "NULL";
        } else {
            throw new IllegalStateException("unsupported data type: " + clz.getName());
        }
    }
}